﻿namespace FakeItEasy.Tests
{
    using System;
    using FakeItEasy.Core;
    using NUnit.Framework;

    [TestFixture]
    public class RaiseTests
    {
        private IFoo foo;
        private object sender;
        private EventArgs eventArguments;

        [SetUp]
        public void Setup()
        {
            this.foo = A.Fake<IFoo>();
            this.foo.SomethingHappened += new EventHandler(this.Foo_SomethingHappened);
            this.sender = null;
            this.eventArguments = null;
        }

        [Test]
        public void Raising_with_sender_and_arguments_should_raise_event_with_specified_sender()
        {
            object senderToUse = new object();
            this.foo.SomethingHappened += Raise.With(senderToUse, EventArgs.Empty).Now;

            Assert.That(this.sender, Is.EqualTo(senderToUse));
        }

        [Test]
        public void Raising_with_sender_and_arguments_should_raise_event_with_specified_arguments()
        {
            var arguments = new EventArgs();

            this.foo.SomethingHappened += Raise.With(this.foo, arguments).Now;

            Assert.That(this.eventArguments, Is.SameAs(arguments));
        }

        [Test]
        public void Raising_with_arguments_only_should_raise_event_with_fake_as_sender()
        {
            this.foo.SomethingHappened += Raise.With(EventArgs.Empty).Now;

            Assert.That(this.sender, Is.SameAs(this.foo));
        }

        [Test]
        public void Raising_with_arguments_only_should_raise_event_with_specified_arguments()
        {
            var arguments = new EventArgs();

            this.foo.SomethingHappened += Raise.With(arguments).Now;

            Assert.That(this.eventArguments, Is.SameAs(arguments));
        }

        [Test]
        public void Now_should_throw_when_called_directly()
        {
            var raiser = new Raise<EventArgs>(null, EventArgs.Empty);

            Assert.Throws<NotSupportedException>(() =>
                raiser.Now(null, null));
        }

        [Test]
        public void Go_should_return_handler_with_Now_as_method()
        {
            Assert.That(Raise.With(EventArgs.Empty).Go.Method.Name, Is.EqualTo("Now"));
        }

        [Test]
        public void WithEmpty_should_return_raise_object_with_event_args_empty_set()
        {
            var result = Raise.WithEmpty();

            var eventArgs = (result as IEventRaiserArguments).EventArguments;

            Assert.That(eventArgs, Is.EqualTo(EventArgs.Empty));
        }

        [Test]
        public void Should_not_fail_when_raising_event_that_has_no_registered_listeners()
        {
            // Arrange
            this.foo = A.Fake<IFoo>();

            // Act

            // Assert
            Assert.That(() => { foo.SomethingHappened += Raise.WithEmpty().Now; }, Throws.Nothing);
        }

        private void Foo_SomethingHappened(object newSender, EventArgs e)
        {
            this.sender = newSender;
            this.eventArguments = e;
        }
    }
}
